.cpu "6502"

;; vic colours
black=$0
white=$1
red=$2
cyan=$3
purple=$4
green=$5
blue=$6
yellow=$7
orange=$8
brown=$9
lred=$a
grey1=$b
grey2=$c
lgreen=$d
lblue=$e
grey3=$f

;; screen memory (almost 4 pages)
screen=$0400
;; colour memory (almost 4 pages)
colours=$d800

;; frame and background colour
frmcol=$d020
bgcol=$d021

;; joystick ports
joy1=$dc01
joy2=$dc00

;; screen dimensions
numlines=25
numcols=40

;; some macros to abbreviate some operations
; move word
movw	.macro
	lda	\2
	sta	\1
	lda	\2+1
	sta	\1+1
	.endm
; move immediate word
movwi	.macro
	lda	#<\2
	sta	\1
	lda	#>\2
	sta	\1+1
	.endm
; move immediate byte
movbi	.macro
	lda	#\2
	sta	\1
	.endm
; add 8 bit immediate to 16 bit word
addwbi	.macro
	clc
	lda	\1
	adc	#\2
	sta	\1
	lda	\1+1
	adc	#0
	sta	\1+1
	.endm
; add 8 bit value to 16 bit word
addwb	.macro
	clc
	lda	\1
	adc	\2
	sta	\1
	lda	\1+1
	adc	#0
	sta	\1+1
	.endm
; sub 8 bit immediate from 16 bit word
subwbi	.macro
	sec
	lda	\1
	sbc	#\2
	sta	\1
	lda	\1+1
	sbc	#0
	sta	\1+1
	.endm
; sub 8 bit value from 16 bit word
subwb	.macro
	sec
	lda	\1
	sbc	\2
	sta	\1
	lda	\1+1
	sbc	#0
	sta	\1+1
	.endm

;; zero page definitions
*=$0

;; temporaries
a:	.byte ?
	.byte ?		; for some reason p doesn't like to be at 1????
p:	.word ?

;; player data
p1a:	.word ?		; address of player 1's paddle
p2a:	.word ?
p1p:	.byte ?		; y position of paddle
p2p:	.byte ?

;; ball data
balla:	.word ?		; address of ball
balllasta:	.word ?	; last address of ball so we can erase
ballx:	.byte ?		; x position
bally:	.byte ?		; y position
ballvx:	.byte ?		; x speed
ballvy:	.byte ?		; y speed

*=$c000

start:
	; init stack
	ldx	$0
	txs

	; set some nice colours
	movbi	bgcol,black
	movbi	frmcol,grey1
	lda	#white
	jsr	fillcolours

	; clear screen
	lda	#$20
	jsr	fillscreen

	; init players
	movwi	p1a,screen+1
	movbi	p1p,0
	movwi	p2a,screen+numcols-2
	movbi	p2p,0

	; init ball speed
	lda	#1
	sta	ballvx
	sta	ballvy

	; draw initial paddles
	lda	#72
	ldx	#0
	jsr	drawpaddle
	lda	#71
	ldx	#$2
	jsr	drawpaddle

	; into the game!
	jmp	newgame


; wait a little while for each frame.
; todo: wait for vsync
wait:
	ldx	#$ff
-	ldy	#$20
-	dey
	bne	-
	dex
	bne	--
	rts

newgame:
	; this is jumped into when we should be returning from a subroutine
	; so just reset the stack here
	ldx	$0
	txs
	jsr	drawball
	; wait a bit between turns
	jsr	wait
	jsr	wait
	jsr	wait
	; set initial ball position (and pointer)
	; speed is unchanged!
	movwi	balla,screen+12*numcols+20
	movbi	ballx,20
	movbi	bally,12
	; fallthrough

;; game loop. process ball, then paddles
mainloop:
	jsr	wait
	jsr	ball
	lda	joy1
	lsr
	bcc	joy1up
	lsr
	bcc	joy1down
gamep2:
	lda	joy2
	lsr
	bcc	joy2up
	lsr
	bcc	joy2down
	jmp 	mainloop

;; handle player 1 input
joy1up:
	ldx	p1p
	beq	gamep2		; paddle can't go up, handle player 2
	dex
	stx	p1p		; decrement position
	ldx	#0
	jsr	clearpaddle	; delete old paddle before drawing new
	ldx	#0
	jsr	moveup		; move the pointer as well
	jmp	paddle1		; and draw the left paddle there
joy1down:
	ldx	p1p
	cpx	#20
	beq	gamep2		; paddle can't go down,handle player 2
	inx
	stx	p1p		; increment position
	ldx	#0
	jsr	clearpaddle
	ldx	#0
	jsr	movedown	; move the pointer down too
paddle1:
	lda	#72		; left paddle character
	ldx	#0
	jsr	drawpaddle
	jmp	gamep2
;; handle player 2 input, same as above
joy2up:
	ldx	p2p		; load position
	beq	mainloop	; can't go up
	dex
	stx	p2p
	ldx	#2
	jsr	clearpaddle
	ldx	#2
	jsr	moveup
	jmp	paddle2
joy2down:
	ldx	p2p
	cpx	#20
	beq	mainloop
	inx
	stx	p2p
	ldx	#2
	jsr	clearpaddle
	ldx	#2
	jsr	movedown
paddle2:
	lda	#71		; right paddle character
	ldx	#$2
	jsr	drawpaddle
	jmp	mainloop

;; draw a paddle with spaces to delete it
clearpaddle:
	lda	#$20
	jmp	drawpaddle

;; move up a paddle pointer
moveup:
	sec
	lda	p1a,x
	sbc	#numcols
	sta	p1a,x
	lda	p1a+1,x
	sbc	#0
	sta	p1a+1,x
	rts
;; move down a paddle pointer
movedown:
	clc
	lda	p1a,x
	adc	#numcols
	sta	p1a,x
	lda	p1a+1,x
	adc	#0
	sta	p1a+1,x
	rts

;; draw a paddle with char in acc
drawpaddle:
	sta	a
	lda	p1a,x
	sta	p
	lda	p1a+1,x
	sta	p+1
	ldy	#0
	ldx	#5
-	lda	a
	sta	(p),y
	addwbi	p,numcols
	dex
	bne	-
	rts

;; process the ball
ball:
	;; first move vertically
	lda	#numcols
	ldx	ballvy
	jsr	ballptr		; move the ball pointer vertically
	clc
	lda	bally
	adc	ballvy
	sta	bally		; and the position
	;; now test collision with top or bottom wall
	;; depending on direction we're moving
	lda	ballvy
	bmi	ballup
balldown:
	lda	bally
	cmp	#numlines-1
	bcs	clampbottom	; branch if acc >=
	jmp	ballhoriz
ballup:
	ldx	bally
	dex
	bmi	clamptop
	; fallthrough

ballhoriz:
	;; next move horizontally
	ldx	ballvx
	lda	#1
	jsr	ballptr		; move the ball pointer horizontally
	clc
	lda	ballx
	adc	ballvx
	sta	ballx		; and the position
	;; test collision.
	;; (nb: since x speed is always 1 or -1 we will never go beyond the paddle!!)
	;; if we hit a paddle, reflect.
	;; otherwise, player loses. this is a bit simplistic perhaps
	lda	ballvx
	bmi	ballleft
ballright:
	lda	ballx
	cmp	#numcols-3
	beq	hitright
	jmp	drawball
flipx:	sec
	lda	#0
	sbc	ballvx
	sta	ballvx
	jmp	drawball
ballleft:
	lda	ballx
	cmp	#2
	beq	hitleft
;; now to draw the ball
drawball:
	ldy	#0
	lda	#$20
	sta	(balllasta),y	; erase the old ball
	lda	#81
	sta	(balla),y	; draw the new ball
	movw	balllasta,balla
	rts

;; we hit a wall, maybe even went beyond it.
;; so reposition the ball to be inside the drawing area
;; and negate its y speed
clamptop:
	;; set ball y to 0, don't change x
	movbi	bally,0
	;; and also set the pointer
	clc
	lda	#<screen
	adc	ballx
	sta	balla
	lda	#>screen
	sta	balla+1
	jmp	flipy
clampbottom:
	;; same as above but set y to numlines-1
	movbi	bally,numlines-1
	clc
	lda	#<screen+(numlines-1)*numcols
	adc	ballx
	sta	balla
	lda	#>screen+(numlines-1)*numcols
	sta	balla+1
	; fallthrough
;; negate y speed cause we hit a wall
flipy:	sec
	lda	#0
	sbc	ballvy
	sta	ballvy
	jmp	ballhoriz

;; check if we hit the left or right paddle
hitleft:
	ldx	#0
	jmp	hitpaddle
hitright:
	ldx	#1
	jmp	hitpaddle

;; y speed for each paddle position
speedy:	.char -2, -1, 0, 1, 2

;; check if he hit a given paddle
hitpaddle:
	lda	bally
	sec
	sbc	p1p,x		; subtract paddle y
	bmi	lose		; ball above paddle
	cmp	#5
	bpl	lose		; below paddle
	; paddle hit the ball, figure out new y speed
	; flip x
	tax
	lda	speedy,x
	sta	ballvy
	jmp	flipx

lose:	jmp	newgame

; move the ball pointer by repeated addition/subtraction
; a: value to add, x: multiplier
ballptr:
	stx	a
	ldx	a	; get flags
	sta	a
	bmi	ballsub
	beq	+
balladd:
	addwb	balla,a
	dex
	bne	balladd
	rts
ballsub:
	subwb	balla,a
	inx
	bne	ballsub
+	rts

;; fill screen with char in acc
fillscreen:
	ldx	#$00
-	sta	screen,x
	sta	screen+$100,x
	sta	screen+$200,x
	sta	screen+$300,x
	inx
	bne	-
	rts

;; fill colours with colour in acc
fillcolours:
	ldx	#$00
-	sta	colours,x
	sta	colours+$100,x
	sta	colours+$200,x
	sta	colours+$300,x
	inx
	bne	-
	rts
